package eye.ss.database;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.EntityManager;

import eye.ss.ServerEnv;
import eye.util.ExceptionUtil;
import eye.util.Modelable;

public interface BaseModel extends Modelable {
	static Models getModels(Class modelType) {
		return ServerEnv.get().getModels(modelType);
	}

	default Query createQuery() {
		return new Query(getClass(), getEm());
	}

	@Override
	default void delete(boolean isAccountDeleting) {
		createQuery().delete(this);
	}

	default EntityManager getEm() {
		return getModels(getClass()).db().getEm();
	}

	default <Z> Z require(String fieldName) {
		try {
			Field f = getClass().getField(fieldName);
			Z value = (Z) f.get(this);
			if (value == null) {
				throw new IllegalStateException(fieldName + " may not be null within " + this);
			}
			return value;
		} catch (SecurityException e) {
			throw ExceptionUtil.wrap(e);
		} catch (NoSuchFieldException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalArgumentException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalAccessException e) {
			throw ExceptionUtil.wrap(e);
		}
	}

	@Override
	default void save() {
		createQuery().save(this);
	}

	default void set(Field f, Long value) throws IllegalArgumentException, IllegalAccessException {
		long id = ((Number) value).longValue();
		Class<? extends BaseModel> type = (Class<? extends BaseModel>) f.getType();
		Models<? extends BaseModel> service = ServerEnv.get().getModels(type);
		Modelable model = service.find(id);
		f.set(this, model);
	}

	@Override
	default void set(String fieldName, Object value) {
		try {

			Field f = getClass().getField(fieldName);

			if (value == null) {
				f.set(this, null);
				return;
			}
			if (BaseModel.class.isAssignableFrom(f.getType())) {
				set(f, ((Number) value).longValue());
			} else if (List.class.isAssignableFrom(f.getType())) {
				setList(f, (List<Number>) value);
			} else {
				f.set(this, value);
			}
		} catch (IllegalArgumentException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalAccessException e) {
			throw ExceptionUtil.wrap(e);
		} catch (SecurityException e) {
			throw ExceptionUtil.wrap(e);
		} catch (NoSuchFieldException e) {
			throw ExceptionUtil.wrap(e);
		}
	}

	default void setList(Field f, List<Number> value) throws IllegalArgumentException, IllegalAccessException {
		ParameterizedType type = (ParameterizedType) f.getGenericType();
		Class<Modelable> t = (Class) type.getActualTypeArguments()[0];
		if (Modelable.class.isAssignableFrom(t)) {
			List<Modelable> children = new ArrayList();
			Models<Modelable> service = getModels(t);
			for (Object item : value) {
				long id = ((Number) item).longValue();
				Modelable model = service.find(id);
				children.add(model);
			}
			f.set(this, children);
		} else {
			f.set(this, value);
		}

	}
}
