package eye.ss.database;

import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityTransaction;
import javax.persistence.TypedQuery;

import org.apache.commons.io.FileUtils;

import com.objectdb.Doctor;

import eye.service.EyeService;
import eye.service.ServiceEnv;
import eye.service.ServiceUtil;
import eye.util.ExceptionUtil;
import eye.util.FileUtil;
import eye.util.StringUtil;
import eye.util.logging.Log;
import eye.util.logging.Log.Cat;

public abstract class DatabaseService extends EyeService {
	static {
		System.setProperty("objectdb.conf", "./conf/db/objectdb.conf");
	}

	private EntityManager em;

	public static DatabaseService get() {
		return ServiceEnv.get().requireService(DatabaseService.class);
	}

	public static boolean isInTransaction() {
		DatabaseService db = ServiceUtil.getServiceIfStarted(DatabaseService.class);
		if (db == null || db.em == null) {
			return false;
		} else {
			return db.em.isOpen() && db.em.getTransaction().isActive();
		}

	}

	protected static File genericGlobalSetup(String dbFlavor, boolean resetDb) {
		Log.config("generic global setup", Cat.DB);

		assert dbFlavor.contains("-") : dbFlavor + " should contain a dash";
		String dir = "./conf/db/" + dbFlavor;
		FileUtil.ensureDir(dir);
		String dbRoot = StringUtil.substring(dbFlavor, null, "-");

		File dbFile = new File(dir + "/" + dbRoot + "-db.odb");

		if (resetDb) {
			Log.info("Deleting old database:" + dbFile.getAbsolutePath());
			FileUtil.delete(dbFile.getAbsolutePath());
			FileUtil.delete(dbFile.getAbsolutePath() + "$");
		} else if (dbFile.exists() == false) {
			File repair = new File(dir + "/" + dbRoot + "-repair.odb");
			if (repair.exists()) {
				Log.warning("Repairing production database. Please delete the repair.odb after you are done", Cat.DB);
				String[] args = { repair.getAbsolutePath(), dbFile.getAbsolutePath() };
				Doctor.main(args);
			} else {
				Log.info("Cannot find " + dbRoot + " so creating it");
			}
		}
		return dbFile;

	}

	public abstract void backup();

	public void checkTransaction(boolean status) {

		if (getTransaction().isActive() != status) {
			if (status) {
				throw new IllegalStateException(
						"You should have an open transaction on " + getTransaction().hashCode());
			} else {
				throw new IllegalStateException("Your transaction should have been committed or rolled back");
			}
		}
	}

	public void commit() {
		EntityTransaction t = getTransaction();
		if (t.isActive()) {
			t.commit();
			Log.fine("Commit on " + getEmf().hashCode(), Cat.DB);
		} else {
			Log.warning("DB commit has no effect", Log.Cat.DB);
		}

	}

	public void ensureClosed() {
		EntityManagerFactory dbEmf = getEmf();
		if (dbEmf != null) {
			if (dbEmf.isOpen()) {
				Log.config("close db", Cat.DB);

				dbEmf.close();
			}
			dbEmf = null;
		}

	}

	public final void ensureTransaction() {
		if (!isInTransaction()) {
			startTransaction();
		}

	}

	public void flush() {
		Log.config("flush", Cat.DB);

		getEm().flush();

	}

	public final EntityManager getEm() {
		if (em == null) {
			em = createEm();
		} else if (em.isOpen() == false) {
			Log.warning("entity manager unexpected closed");
			try {
				em = createEm();
			} catch (Throwable e) {
				Log.warning("Hail Mary restart");
				reconnect();
			}
		}
		return em;
	}

	public EntityTransaction getTransaction() {
		return getEm().getTransaction();
	}

	@Override
	public boolean isCritical() {
		return true;
	}

	public void maybeCommit() {
		EntityTransaction t = getTransaction();
		if (t.isActive()) {
			commit();

		}
	}

	public void rollBack() {
		if (ServiceEnv.get() != null) {
			EntityTransaction transaction = getTransaction();
			if (transaction.isActive()) {
				transaction.rollback();
				Log.config("Rollback Transation on " + getEmf().hashCode(), Cat.DB);
				assert transaction.isActive() == false;
			}

		} else {
			Log.warning("Had no env when request for rollBack came in");
		}
	}

	public void startTransaction() {
		getEm().getTransaction().begin();
		Log.fine("Start Transaction on " + getEmf().hashCode(), Cat.DB);

	}

	protected void backup(String dir) {
		Log.config("Backup " + dir, Cat.DB);
		TypedQuery<Thread> backupQuery = getEm().createQuery("objectdb backup", Thread.class);
		File backupDir = FileUtil.ensureEmptyDir(dir + "/backup-dir");
		backupQuery.setParameter("target", backupDir);
		Thread backupThread = backupQuery.getSingleResult();
		try {
			backupThread.join();
			List<File> backup = FileUtil.getFiles(backupDir, true, "odb");
			assert backup.size() == 1 : backup + " should have one file";
			FileUtils.copyFile(backup.get(0), new File(dir + "/backup.db.odb"));
		} catch (InterruptedException | IOException e) {
			throw ExceptionUtil.wrap(e);
		}
	}

	@Override
	protected void doClose() {
		super.doClose();
		if (em != null && em.isOpen()) {
			Log.config("close", Cat.DB);
			em.close();
		}
	}

	protected abstract EntityManagerFactory getEmf();

	@Override
	protected void init() {

	}

	protected abstract void reconnect();

	private final EntityManager createEm() {
		assert isReady();
		Log.config("create entity manager", Cat.DB);

		assert getEmf() != null;
		return getEmf().createEntityManager();
	}

}
