package eye.ss.database;

import java.util.List;
import java.util.logging.Level;

import javax.persistence.EntityManager;
import javax.persistence.TypedQuery;

import eye.util.Modelable;
import eye.util.StringUtil;
import eye.util.logging.Log;
import eye.util.logging.Log.Cat;
import models.Account;

public class Query<E extends Modelable> {

	private static boolean logQuery = false;
	private final Class<E> clazz;
	private final EntityManager em;
	private final StringBuilder accum = new StringBuilder();
	private boolean readyToFlush;
	public boolean allowSubtypes;

	public Query(Class<E> clazz, EntityManager em) {
		this.clazz = clazz;
		this.em = em;
		assert em.isOpen() : "Entity manger should be open";

	}

	public static <E extends Modelable> Query<E> create(Class<E> clazz, EntityManager em) {
		return new Query(clazz, em);
	}

	public static void logQuery(String name) {
		logQuery = true;
		Log.setLevel(Cat.DB, Level.CONFIG);
		Log.config("--------------------Logging Query" + name + "--------------------", Cat.DB);
	}

	public static void stopLogging() {
		if (logQuery) {
			Log.setLevel(Cat.DB, Log.getDefaultLevel());
			logQuery = false;
		}
	}

	public String computeUniqueName(String name) {
		if (name == null) {
			name = "Unnamed";
		}
		name = StringUtil.getPrefix(name, "-");
		List<E> values = findNameCousins(name);

		int max = 0;
		for (E value : values) {
			int index = value.getName().lastIndexOf("-");
			String indexValue = value.getName().substring(index + 1);

			int intValue = Integer.parseInt(indexValue);

			max = Math.max(max, intValue);
		}

		String value = name + "-" + (max + 1);
		return value;
	}

	public int count() {
		emitCount();
		Number result = createObject(Number.class);
		return result.intValue();
	}

	public void delete(E child) {
		em.remove(child);
	}

	public List<E> findAll() {
		emitSelect(false);
		List<E> values = createList();
		return values;
	}

	public List<E> findAll(int offset, int max) {
		emitSelect(false);
		List<E> values = createList(offset, max);
		return values;
	}

	public List<E> findAllByStringField(String fieldName, String fieldValue, int offset, int max) {
		emitSelect();
		emit("target." + fieldName + "='" + fieldValue + "'");

		List<E> answers = createList(offset, max);
		return answers;
	}

	public List<E> findAllNonNull(String fieldName) {
		emitSelect();
		emitNotNull(fieldName);
		List<E> values = createList();
		return values;
	}

	public E findById(long id) {
		emitSelect();
		emitId(id);
		return createChild();
	}

	public E findByLabel(String label) {
		return findByString("label", label);

	}

	public E findByLabel(String label, String ownerField, Modelable parent) {
		emitSelectChildren(ownerField, parent);
		emit("AND");
		//HACK
		label = label.replace("'", "''");
		emit("target.label='" + label + "'");
		E e = createChild();
		return e;
	}

	public E findByName(String name) {
		emitSelect();
		emitName(name);
		return createChild();
	}

	public E findByName(String name, String ownerField, Modelable parent) {
		emitSelectChildren(ownerField, parent);
		emit("AND");
		emitName(name);
		return createChild();
	}

	public E findByNameIgnoreCase(String name) {
		emitSelect();
		emitNameIgnoreCase(name);
		E result = createChild();
		return result;
	}

	public List<E> findByOwner(String ownerField, Modelable parent) {
		emitSelectChildren(ownerField, parent);
		return createList();
	}

	public List<E> findByOwner(String ownerField, Modelable... parents) {
		emitSelectChildren(ownerField, parents);
		return createList();
	}

	public List<E> findByOwner(String ownerField, Modelable owner, String sortBy, String query) {
		emitSelect();
		emitSelectChildren(ownerField, owner);
		emit(" AND ");
		emit(query);
		emit(" ORDER BY target.");
		emit(sortBy);
		return createList();
	}

	public List<E> findByOwnerAndClause(String ownerField, Modelable parent, String clause) {
		emitSelectChildren(ownerField, parent);
		emit("AND");
		emit(clause);
		return createList();
	}

	public List<E> findByOwnerWithQuery(String ownerField, Modelable parent, String query) {

		emitSelectChildren(ownerField, parent);
		emit(" AND ");
		emit(query);
		return createList();
	}

	public E findByString(String key, String value) {
		emitSelect();
		emit("target." + key + "='" + value + "'");
		return createChild();
	}

	public E findByStringField(String fieldName, String fieldValue) {
		emitSelect();
		emit("target." + fieldName + "='" + fieldValue + "'");

		List<E> answers = createList();
		if (answers.size() > 1) {
			Log.warning(fieldValue + " should have had only once answer. It Has:" + answers);
		}
		if (answers.size() == 1) {
			return answers.get(0);
		} else {
			return null;
		}

	}

	public E findBySubstring(String label, Account account) {
		emitSelectChildren("account", account);
		emit("AND LOCATE(target.label, '" + label + "')>0");
		return createChild();
	}

	public List<E> findListByString(String key, String value) {
		emitSelect();
		emit("target." + key + "='" + value + "'");
		return createList();
	}

	public boolean isLabelUnique(String name) {
		emitSelect();
		emitLabel(name);
		return size() == 0;
	}

	public boolean isNameUnique(Modelable parent, String ownerField, String name) {
		emitSelectChildren(ownerField, parent);
		emit("AND");
		emitName(name);
		int size = size();
		return size == 0;
	}

	public boolean isNameUnique(String name) {
		emitSelect();
		emitName(name);
		return size() == 0;
	}

	public void save(E modelable) {
		modelable.validate();

		if (!em.contains(modelable)) {
			em.persist(modelable);
		}
		assert modelable.getId() != 0;
	}

	@Override
	public String toString() {
		return "Query:" + accum;
	}

	private E createChild() {
		try {
			TypedQuery<E> query = createQuery();
			readyToFlush = true;
			List<E> list = query.getResultList();
			if (list.size() == 0) {
				return null;
			} else if (list.size() > 1) {
				Log.warning("While processing " + query + " Expected 1 result, got:" + list);
				return list.get(0);
			} else {
				return list.get(0);
			}

		} catch (Exception e) {

			throw new RuntimeException("While processing \n\t" + accum + "\n" + " in " + em, e);
		} finally {
			stopLogging();
		}
	}

	private List<E> createList() {
		try {

			TypedQuery<E> query = createQuery();

			List<E> e = query.getResultList();

			readyToFlush = true;
			Log.config("Finished creating list", Cat.DB);
			return e;
		} catch (Exception e) {
			Log.fatal("while processing" + accum, e);
			throw new RuntimeException("While processing \n\t" + accum + "\n caught expection:\n\t" + e.getMessage(),
					e);
		} finally {
			stopLogging();
		}
	}

	private List<E> createList(int offset, int max) {
		try {

			TypedQuery<E> query = createQuery();
			query.setFirstResult(offset).setMaxResults(max).getResultList();
			List<E> e = query.getResultList();
			readyToFlush = true;
			Log.config("Finished creating list", Cat.DB);
			return e;
		} catch (Exception e) {
			Log.fatal("while processing" + accum, e);
			throw new RuntimeException("While processing \n\t" + accum + "\n caught expection:\n\t" + e.getMessage(),
					e);
		} finally {
			stopLogging();
		}
	}

	private <X extends Object> X createObject(Class<X> type) {
		try {
			TypedQuery<X> query = em.createQuery(accum.toString(), type);
			return query.getSingleResult();

		} catch (Exception e) {
			throw new RuntimeException("While processing \n\t" + accum + "\n caught expection:\n\t" + e.getMessage(),
					e);
		} finally {
			stopLogging();
		}
	}

	private TypedQuery<E> createQuery() {
		String result = accum.toString();
		Log.config(result, Cat.DB);
		TypedQuery<E> query = em.createQuery(accum.toString(), clazz);

		Log.fine("set lazy", Cat.DB);

		return query;
	}

	private Query emit(String str) {
		if (readyToFlush) {
			accum.setLength(0);
			readyToFlush = false;
		}
		accum.append(str);
		accum.append(" ");
		return this;
	}

	private void emitCount() {
		emit("SELECT Count(target) from " + clazz.getSimpleName() + " target");
	}

	private void emitId(long id) {
		emit("target.id=" + id);
	}

	private void emitIn(String ownerField, Modelable... parents) {
		emit("(");
		String target = "target." + ownerField + ".id =";
		for (int i = 0; i < parents.length; i++) {
			if (i != 0) {
				emit(" OR ");
			}
			emit(target + parents[i].getId());
		}
		emit(")");
	}

	private void emitLabel(String name) {
		emit("target.label='" + name + "'");
	}

	private void emitName(String name) {
		emit("target.name='" + name + "'");
	}

	private void emitNameIgnoreCase(String name) {
		name = name.toLowerCase();
		emit("LOWER(target.name)='" + name + "'");

	}

	private void emitNotNull(String fieldName) {
		emit("target." + fieldName + " IS NULL");
	}

	private void emitSelect() {
		emitSelect(true);
	}

	private void emitSelect(boolean includeAnd) {
		emit("SELECT target from " + clazz.getSimpleName() + " target WHERE");
		if (!allowSubtypes) {
			emit("TYPE(target)=" + clazz.getSimpleName());
			if (includeAnd) {
				emit("AND ");
			}
		}
	}

	private void emitSelectChildren(String ownerField, Modelable... parents) {
		emitSelect();
		emitIn(ownerField, parents);
	}

	private List<E> findNameCousins(String name) {
		emitSelect();
		emit("target.name LIKE '" + name + "-%'");
		return createList();
	}

	private int size() {
		try {
			Log.fine("get size", Cat.DB);
			return em.createQuery(accum.toString(), clazz).getResultList().size();
		} catch (Exception e) {
			throw new RuntimeException("While processing \n\t" + accum + "\n caught expection:\n\t" + e.getMessage(),
					e);
		}
	}

}
