package models;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import eye.play.control.PlayControl;
import eye.ss.database.Models;
import eye.ss.database.Query;
import eye.transfer.EyeRecord;
import eye.transfer.EyeTable;
import eye.transfer.EyeType;
import eye.util.DateUtil;
import eye.util.EncryptUtil;
import eye.util.UserException;
import eye.util.collection.ListUtil;
import eye.util.logging.Log;

public class Accounts extends PlayModels<Account> {

	public EyeTable createHome(Account account) {
		EyeTable t = Models.createModelTable();

		t.setName("home");

		List<Account> accounts = new ArrayList();
		accounts.add(account);

		if (PlayControl.getEnv().superAdminMode) {
			if (account.isAdmin()) {
				accounts.addAll(findAll());
			} else {
				if (account.adminForMouseGroups != null) {
					for (String group : account.adminForMouseGroups) {
						List<Account> mouseGroups = getByGroup(group);
						Collections.sort(mouseGroups);
						ListUtil.addUnion(accounts, mouseGroups);
					}
				}
				ListUtil.addUnion(accounts, account);
				accounts.add(PlayControl.global);
			}
		} else {

			if (account.equals(PlayControl.global) == false) {
				// hack
				PlayControl.global.label = "System";
				accounts.add(PlayControl.global);
			}
			if (account.equals(PlayControl.test) == false) {
				accounts.add(PlayControl.test);
			}

		}

		load(t, PlayControl.courses, account, "Taught Courses", PlayControl.courses.getModels(account));
		if (account.enrolled != null) {
			load(t, PlayControl.courses, account, "Courses", account.enrolled);
		}

		for (Account target : accounts) {
			load(t, PlayControl.watchlists, target);
			load(t, PlayControl.pickFilters, target);
			load(t, PlayControl.tradingModels, target);
			load(t, PlayControl.macros, target);
		}
		return t;
	}

	@Override
	public final Account find(EyeRecord record) {
		return require(record.getName());
	}

	public Account get(String name) {
		Query<Account> q = createQuery();
		Account strat = q.findByName(name);
		if (strat == null && name.contains("@")) {
			strat = q.findByStringField("email", name);
		}

		if (strat == null) {
			strat = q.findByNameIgnoreCase(name);
		}

		return strat;

	}

	public List<Account> getByGroup(String group) {
		return createQuery().findListByString("mouseGroup", group);
	}

	@Override
	public EyeType getEyeType() {
		return EyeType.account;
	}

	@Override
	public EyeRecord getModels() {
		EyeRecord r = new EyeRecord();
		r.setName("availableRecords");
		List<Account> accounts = PlayControl.accounts.findAll();
		EyeTable table = getModels(accounts);
		r.add(table);
		return r;
	}

	public EyeTable getModels(List<Account> accounts) {
		EyeTable table = new EyeTable();
		table.setName("listings");
		table.addColumnTypes(EyeType.String, EyeType.String, EyeType.String, EyeType.Boolean, EyeType.String,
				EyeType.Date, EyeType.String);
		table.addColumnNames("label", "name", "e-mail", "send email", "org", "last login", "password");

		Collections.sort(accounts);
		for (Account account : accounts) {
			int size = 7;

			Object[] row = new Object[size];
			row[0] = account.getLabel();
			row[1] = account.name;
			row[2] = account.email;
			row[3] = account.sendEmail;
			row[4] = account.mouseGroup;
			if (row[4] == null) {
				row[4] = account.org;
			}

			if (account.lastLogin != null) {
				row[5] = DateUtil.serverFormat.format(account.lastLogin);
			}
			row[6] = account.password;
			table.addRow(row);
		}
		return table;
	}

	@Override
	public EyeRecord load(Account target) {
		EyeRecord record = super.load(target);
		record.set("recordAccount", target.name);
		if (target.shadowLogins != null) {
			record.set("encryptedPassword", EncryptUtil.encrypt(target.password));
		}
		return record;
	}

	public final Account require(String name) {
		Account account = get(name);
		if (account == null) {
			throw new UserException("The account with the name " + name + " cannot be found.", false);
		}

		return account;
	}

	@Override
	protected Account bind(Account account, EyeRecord r) {
		super.bind(account, r);
		r.bind(account, "sendEmail");
		r.bind(account, "tearSheet");
		r.bind(account, "brokerage");

		if (r.updateModel == false) {
			r.bind(account, "acceptedTerms");
			r.bind(account, "isMoneyManager");
			r.bind(account, "brokerageAccount");
			r.set("isStudent", account.isStudent());

			r.set("isProfessor", account.isProfessor());
		}
		if (r.isClientTarget()) {
			r.bind(account, "status");
		}
		r.maybeBind(account, "studentId");
		return account;
	}

	@Override
	protected Account create() {
		// Dont create accounts this way.
		return null;
	}

	protected void load(EyeTable t, AccountMembers service, Account target, String label, List strats) {
		List<Strategy> stratList = strats;
		if (ListUtil.isEmpty(stratList)) {
			return;
		}

		String summary = "A list of " + service.getEyeType().label().toLowerCase() + "s ";
		if (target == PlayControl.global) {
			summary += "defined by Equities Lab for general usage";
			label = "System " + label;
		} else if (target == PlayControl.test) {
			summary += " for use as examples";
			label = "Example " + label;
		} else if (PlayControl.getEnv().account == target) {
			summary += " in your account";
			label = "My " + label;
		} else {
			summary += " in " + target.description;
			label = target.label;
		}
		// There is no account here, so cannot use normal eye table.
		service.sort(stratList);
		Object[] summaryRow = { label, summary, null, -1l, service.getEyeType() };
		t.addRow(summaryRow);
		boolean validCourse = false;
		for (Strategy strat : stratList) {
			//Reference to dead course
			if (strat.label == null) {
				continue;
			}
			validCourse = true;
			if (strat.description.contains("\n") || strat.description.contains("\t")) {
				Log.severe("Description contains \n or \t:" + strat + "\n description:" + strat.description);
				strat.description = "Sorry, there was a error in your description";
			}
			Object[] row = { strat.getLabel(), strat.description, label, strat.id, service.getEyeType() };
			t.addRow(row);
		}
		if (validCourse == false) {
			t.getRows().remove(summaryRow);
		}
	}

	private void load(EyeTable t, AccountMembers service, Account target) {
		List<Strategy> stratList = service.getModels(target);
		if (stratList.size() == 0) {
			return;
		}
		String label = service.getEyeType().label() + "s";
		load(t, service, target, label, stratList);
	}

}
