package models;

import static eye.EyeConstants.*;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map.Entry;

import eye.play.control.PlayControl;
import eye.transfer.EyeData;
import eye.transfer.EyeRecord;
import eye.transfer.EyeTable;
import eye.transfer.EyeType;
import eye.util.StringUtil;
import eye.util.UserException;
import eye.util.collection.ListUtil;
import eye.util.logging.Log;

public class Courses extends AccountMembers<Course> {

	public EyeRecord addCourse(Course course, Account account) {

		if (course.members == null) {
			course.members = new ArrayList();
		}

		if (course.members.contains(account)) {
			throw new UserException("You already have joined " + course.toHandle(), true);
		}
		course.members.add(account);
		if (account.enrolled == null) {
			account.enrolled = new ArrayList();
		} else {
			for (int i = 0; i < account.enrolled.size(); i++) {
				Course myCourse = account.enrolled.get(i);
				if (myCourse.members == null) {
					account.enrolled.remove(i);
					--i;
				}
			}
		}
		account.enrolled.add(course);
		EyeRecord r = new EyeRecord();
		r.setModel(course);
		return r;
	}

	public EyeRecord addCourse(String courseId, Account account) {
		Course course = getGlobal(courseId);

		if (course == null) {
			throw new UserException("<HTML>Cannot find '<b>" + courseId + "</b>' course", true);
		}
		return addCourse(course, account);
	}

	public Course create(String name) {
		checkUniqueName(name);
		Course newGroup = super.create();
		newGroup.name = name;
		newGroup.account = PlayControl.getEnv().account;
		newGroup.description = "Description of <b>" + name + "</b>";
		newGroup.label = "My Course";
		if (name.contains(".")) {
			String suffix = name.substring(name.lastIndexOf(".") + 1);
			if (StringUtil.hasContent(suffix)) {
				newGroup.label = suffix.toUpperCase() + " Course";
			}
		}
		newGroup.lastModified = new Date();
		newGroup.save();

		return newGroup;
	}

	@Override
	public Course delete(long id) {
		Course child = find(id);

		Account owner = PlayControl.getEnv().getAccount();
		Account other = ((Strategy) child).account;
		if (other.equals(owner) == false) {
			removeMember(child, owner);
		} else {
			child.delete(false);
		}
		return child;
	}

	public void delete(String sem) {
		Course group = get(sem);
		if (group != null) {
			group.delete(false);
		}
	}

	@Override
	public Course ensure(EyeRecord r) {
		Long id = r.getId();
		Course strat = null;
		if (id == null) {
			String name = r.getName();
			strat = create(name);
		} else {
			strat = createQuery().findById(id);

			assert strat != null : "There is no strategy with the id of " + id;
		}
		return strat;
	}

	@Override
	public EyeType getEyeType() {
		return EyeType.course;
	}

	public List<PickFilter> getHomeworksForImport(boolean onlyPublic) {
		List<PickFilter> homeworks = new ArrayList();
		Account account = PlayControl.getEnv().account;
		for (Course c : findAll()) {
			if (c.homeworks == null) {
				continue;
			}
			boolean mine = account.equals(c.account);
			for (PickFilter homework : c.homeworks) {
				if (mine) {
					homeworks.add(homework);
				} else {
					if (onlyPublic == false || homework.getMeta("meta-public-homework", true)
							&& !homework.getMeta("meta-hide-homework", false)) {
						homeworks.add(homework);
					}

				}

			}
		}
		return homeworks;
	}

	public EyeRecord getHomeworksForImportRecord() {
		List<PickFilter> homeworks = getHomeworksForImport(true);

		EyeTable table = createModelTable();

		for (PickFilter homework : homeworks) {
			addToTable(table, homework, homework.account.getLabel(), EyeType.homework);
		}
		EyeRecord rec = new EyeRecord();
		rec.add(table);
		return rec;
	}

	@Override
	public EyeRecord getModels() {
		List<Course> courses = findAll();
		EyeRecord r = new EyeRecord();
		EyeTable table = createModelTable();
		table.addColumn("name", EyeType.String);
		for (Course course : courses) {
			if (course.account.isAdmin()) {
				addToTable(table, course, course.account.getLabel(), EyeType.course, course.getName());
			}
		}
		r.add(table);
		return r;
	}

	public Double getScore(PickFilter homework) {
		return (Double) homework.getMetadata().get(META_HOMEWORK_GRADE);
	}

	@Override
	public EyeRecord load(Course strat) {
		EyeRecord record = super.load(strat);
		record.includeNull = false;
		record.set("recordAccount", strat.account.name);
		record.set("course-name", strat.name);
		if (strat.label == null) {
			record.setLabel("New Title");
		}
		if (strat.description == null) {
			record.setDescription("Description of <b>" + strat.name + "</b>. ");
		}

		record.set("homeworks", strat.createHomeworkTable());
		if (strat.isBrowsing() == false) {
			record.set("student-table", strat.createStudentTable());
		}
		return record;
	}

	public void removeMember(Course course, Account removeMe) {
		boolean removed = course.members.remove(removeMe);
		assert removed : course.members + " did not contain " + removeMe;
		removed = removeMe.enrolled.remove(course);
		assert removed : removeMe + " was not enrolled in " + course;
		if (course.homeworks != null) {
			for (PickFilter homework : course.homeworks) {
				List<PickFilter> filters = PlayControl.pickFilters.getMyOwned(homework.id);
				while (ListUtil.hasContent(filters)) {
					filters.remove(0).delete(false);
				}
			}
		}
	}

	public EyeRecord workOnHomework(Long homeworkId) {
		List<PickFilter> owned = PlayControl.pickFilters.getMyOwned(homeworkId);
		PickFilter master = PlayControl.pickFilters.find(homeworkId);

		if (ListUtil.isEmpty(owned)) {
			PickFilter myHomework = master.copy();
			myHomework.master = master;
			myHomework.save();
			return PlayControl.pickFilters.load(myHomework);
		}
		if (owned.size() > 1) {
			Log.warning(owned + " should never be bigger then one. Id was " + homeworkId);
		}
		PickFilter myHomework = owned.get(0);
		if (isSubmitted(myHomework)) {
			return PlayControl.pickFilters.load(myHomework);
		}

		//Updating an in-process homework;
		for (Entry<String, Object> entry : master.getMetadata().entrySet()) {
			myHomework.getMetadata().put(entry.getKey(), entry.getValue());
		}
		myHomework.rebalance = master.rebalance;
		myHomework.description = master.description;
		myHomework.label = master.label;
		EyeRecord result = PlayControl.pickFilters.load(myHomework);
		EyeData myStrat = result.get("strategy");
		EyeData masterStrat = master.getStrategy();
		for (Entry<String, Object> value : masterStrat.getEntries()) {
			if (value.getKey().startsWith("var-grading_")) {
				myStrat.set(value.getKey(), value.getValue());

			}
		}
		return result;

	}

	@Override
	protected Course bind(Course strat, EyeRecord r) {
		super.bind(strat, r);
		bindMeta(strat, r);
		if (r.updateModel) {
			strat.lastModified = new Date();
		}
		return strat;
	}

	@Override
	protected void setAccount(Course child, Account account) {
		child.account = account;
	}

	@Override
	protected void setName(Course child, String name) {
		child.name = name;
	}

	private void checkUniqueName(String name) {
		if (!createQuery().isNameUnique(name)) {
			throw new UserException(name + " has already been used", true);
		}

	}

	private boolean isSubmitted(PickFilter myHomework) {
		return getScore(myHomework) > 0;
	}

}
