package models;

import java.util.Date;
import java.util.List;

import eye.play.control.PlayControl;
import eye.transfer.EyeRecord;
import eye.transfer.EyeTable;
import eye.util.DateUtil;
import eye.util.StopWatch;
import eye.util.StringUtil;
import eye.util.UserException;

public abstract class Editors<GChild extends Strategy> extends AccountMembers<GChild> {

	public final EyeTable createModelTable(List<GChild> strats) {
		EyeTable table = createModelTable();
		sort(strats);
		for (GChild strat : strats) {
			// We only want strategies associated with tactics here.
			if (isAvailable(strat)) {
				assert strat.getLabel() != null : strat.getClass().getSimpleName() + " "
						+ " should not have a null label with id " + strat.id;
				if (strat.account.getLabel().equals("Global")) {
					throw new IllegalStateException("Which account has Global as a label..." + strat.account);
				}
				addToTable(table, strat, strat.account.getLabel() + "'s", getEyeType());
			}
		}
		return table;
	}

	@Override
	public GChild delete(long id) {
		validateDelete(id);
		return super.delete(id);

	}

	@Override
	public EyeRecord getModels() {
		StopWatch watch = new StopWatch("available records");

		List<GChild> strats = findModels();

		EyeRecord r = new EyeRecord();
		r.setName("availableRecords");
		watch.report("got list");

		EyeTable table = createModelTable(strats);

		r.add(table);
		watch.report("done");
		return r;
	}

	public EyeRecord loadByLabel(String label) {
		GChild child = findByLabel(label);
		if (child == null) {
			throw new UserException("Cannot find " + label, false);
		}
		return load(child);

	}

	public EyeRecord prepareForBacktest(Long id, Date begin, Date end) {
		EyeRecord r = prepareForUpdate(id);
		r.set("updateArgs", "backtest");
		if (begin != null) {
			r.set("backtestBegin", begin);
		}
		if (end == null) {
			end = DateUtil.getPureDate();
		}
		r.set("backtestEnd", end);
		return r;
	}

	@Override
	protected GChild bind(GChild strat, EyeRecord r) {

		Account myAccount = PlayControl.getEnv().account;
		if (r.updateModel) {
			if (myAccount != strat.account && PlayControl.getEnv().superAdminMode == false) {
				strat = strat.copy();
				strat.copied = true;
			}
			strat.lastModified = new Date();
		} else {
			assert strat.account != null;
			r.set("recordAccount", strat.account.name);
		}
		super.bind(strat, r);
		r.bindYaml(strat, "content", "strategy");
		if (r.updateModel) {
			assert StringUtil.notEmpty(strat.content);
		}
		return strat;
	}

	protected EyeRecord prepareForUpdate(Long id) {
		EyeRecord r = load(id);
		return r;
	}

	@Override
	protected void setAccount(GChild child, Account account) {
		child.setAccount(account);
	}

	private void validateDelete(long id) {
		GChild child = find(id);
		Account currentAccount = PlayControl.getEnv().getAccount();
		Account stratAccount = ((Strategy) child).account;
		if (stratAccount.equals(currentAccount)) {
			return;
		}

		//Professor allowed to delete student homeworks.
		if (child.master instanceof Strategy) {
			Strategy master = (Strategy) child.master;
			Account masterOwner = master.account;
			if (currentAccount.equals(masterOwner)) {
				return;
			}
		}

		//If you are in super admin mode, you may delete at will.
		if (PlayControl.getEnv().superAdminMode) {
			return;
		}

		throw new UserException(
				"You cannot delete " + child.getLabel() + " as it's owned by " + stratAccount.getLabel(), false);

	}

}
