package models;

import java.lang.reflect.Field;
import java.util.List;

import javax.persistence.Cacheable;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.SequenceGenerator;

import eye.ss.database.BaseModel;
import eye.ss.database.Query;
import eye.util.ExceptionUtil;
import eye.util.ObjectUtil;
import eye.util.ReflectUtil;
import eye.util.UserException;
import eye.util.logging.Log;

@SequenceGenerator(name = "seq", initialValue = 1, allocationSize = 100)
@Inheritance(strategy = InheritanceType.TABLE_PER_CLASS)
@Entity
@Cacheable
public abstract class EyeModel implements Comparable<EyeModel>, BaseModel {

	public String label;

	@Column(columnDefinition = "VARCHAR(2000)")
	public String description;

	@GeneratedValue(strategy = GenerationType.SEQUENCE, generator = "seq")
	@Id
	public long id;
	public Integer rank = Integer.valueOf(0);

	@Override
	public int compareTo(EyeModel model) {
		int result = compareBody(model);
		assert result == -1 * model.compareBody(this) : this + " and " + model + " do not compare correctly";
		return result;
	}

	public <X extends EyeModel> X copy() {
		return (X) ReflectUtil.copy(this, "id", "__");
	}

	@Override
	public Query createQuery() {
		return new Query(this.getClass(), DB.getEm());
	}

	@Override
	public void delete(boolean isAccountDeleting) {
		createQuery().delete(this);
	}

	@Override
	public <Z> Z get(String fieldName) {
		try {
			Field f = this.getClass().getField(fieldName);
			Object value = f.get(this);
			return (Z) value;
		} catch (SecurityException e) {
			throw ExceptionUtil.wrap(e);
		} catch (NoSuchFieldException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalArgumentException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalAccessException e) {
			throw ExceptionUtil.wrap(e);
		}
	}

	@Override
	public String getDescription() {
		return description;
	}

	@Override
	public long getId() {
		return id;
	}

	@Override
	public final String getLabel() {
		return label;
	}

	@Override
	public String getName() {
		return getClass().getSimpleName() + "_" + getId();
	}

	public abstract EyeModel getOwner();

	public void moveTo(Account target) {
		throw new UserException("Sorry, move to is not implemented for " + this, false);
	}

	@Override
	public <Z> Z require(String fieldName) {
		try {
			Field f = this.getClass().getField(fieldName);
			Z value = (Z) f.get(this);
			if (value == null) {
				throw new IllegalStateException(fieldName + " may not be null within " + this);
			}
			return value;
		} catch (SecurityException e) {
			throw ExceptionUtil.wrap(e);
		} catch (NoSuchFieldException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalArgumentException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalAccessException e) {
			throw ExceptionUtil.wrap(e);
		}
	}

	@Override
	public void save() {
		createQuery().save(this);
	}

	@Override
	public void set(String fieldName, Object value) {
		try {

			Field f = this.getClass().getField(fieldName);

			if (value == null) {
				f.set(this, null);
				return;
			}
			if (EyeModel.class.isAssignableFrom(f.getType())) {
				set(f, ((Number) value).longValue());
			} else if (List.class.isAssignableFrom(f.getType())) {
				setList(f, (List<Number>) value);
			} else {
				f.set(this, value);
			}
		} catch (IllegalArgumentException e) {
			throw ExceptionUtil.wrap(e);
		} catch (IllegalAccessException e) {
			throw ExceptionUtil.wrap(e);
		} catch (SecurityException e) {
			throw ExceptionUtil.wrap(e);
		} catch (NoSuchFieldException e) {
			throw ExceptionUtil.wrap(e);
		}
	}

	@Override
	public void setDescription(String description) {
		this.description = description;
	}

	@Override
	public void setLabel(String label) {
		this.label = label;
	}

	@Override
	public String toHandle() {
		if (label != null) {
			return getLabel();
		} else {
			return getName();
		}

	}

	@Override
	public String toString() {
		return getLabel();
	}

	@Override
	public void validate() {
	}

	private int compareBody(EyeModel model) {
		assert model != null : "Why is " + this + " being compared with a null object?";
		assert rank != null;
		assert model.rank != null;
		if (ObjectUtil.notEquals(rank, model.rank)) {
			// Higher ranges
			return Integer.compare(model.rank, rank);
		}

		String label = getLabel();
		String otherLabel = model.getLabel();
		if (label == null) {
			Log.warning("why does " + this + " have no label?");
			return 0;
		}
		if (otherLabel == null) {
			Log.warning("Why does " + model + " have no label?");
			return 0;
		}

		boolean labelInCat = label.contains(":");
		boolean otherInCat = otherLabel.contains(":");

		// If either both or neither in categories, we use the label.
		if (labelInCat == otherInCat) {
			return label.compareTo(otherLabel);
		} else if (labelInCat) {
			// When we have a label and it doesn't, we go first.
			return -1;
		} else {
			// When they have a label we go after them.
			return 1;
		}
	}

}
