package models;

import java.util.Date;
import java.util.List;

import javax.jdo.annotations.Index;
import javax.persistence.Entity;
import javax.persistence.ManyToOne;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

import eye.service.stock.Signal;
import eye.service.stock.SignalReason;
import eye.util.ExceptionUtil;
import eye.util.UserException;
import eye.util.logging.Log;

/**
 * In order to play with the trading model, the stock pick can customize it here.
 *
 * @author Emily
 *
 */
@Entity
public class StockPick extends AccountMember {
	// Watchlist
	@ManyToOne

	@Index
	public Watchlist watchlist;

	// Denormalized for ease of use.
	@ManyToOne

	@Index
	public Account account;

	public String security;

	// Pending actions are Buy, Sell, or Error...
	public String signal = "out";

	public String[] signalCause;

	@Deprecated
	public List<Date> bought;
	@Temporal(TemporalType.DATE)
	public Date signalDate;

	@Temporal(TemporalType.DATE)
	public Date signalStartDate;

	@Temporal(TemporalType.DATE)
	public Date signalEndDate;

	@Temporal(TemporalType.DATE)
	public Date start;

	@Temporal(TemporalType.DATE)
	public Date end;

	public StockPick() {
	}

	public void becomeHold() {
		if (start == null) {
			start = new Date();
		}
		updateSignal(Signal.hold);

	}

	public void becomeOut() {
		if (start == null) {
			delete(false);
		} else {
			int days = (int) ((start.getTime() - new Date().getTime()) / 86400000);
			if (days < 2) {
				delete(false);
			} else {
				if (end == null) {
					end = new Date();
				}
				updateSignal(Signal.out);
			}
		}
	}

	public void clearPending() {
		assert isPending();
		Signal sig = Signal.parse(signal);
		switch (sig) {
		case buy:
			assert getStart() == null;
			delete(false);
			break;
		case sell:
			signal = Signal.hold.name();
			break;
		default:
			throw ExceptionUtil.throwUnsupported(sig);
		}
		signalDate = null;
		signalCause = null;
	}

	@Override
	public int compareTo(EyeModel model) {

		StockPick other = (StockPick) model;
		if (!other.watchlist.equals(watchlist)) {
			return watchlist.compareTo(other.watchlist);
		} else {
			if (!signal.equals(other.signal)) {
				return Signal.parse(signal).compareTo(Signal.parse(other.signal));
			}
			return security.compareTo(other.security);
		}
	}

	@Override
	public <X extends EyeModel> X copy() {
		StockPick copy = super.copy();
		copy.start = null;
		copy.end = null;

		return (X) copy;
	}

	@Override
	public void delete(boolean isAccountDeleting) {
		if (watchlist != null) {
			if (watchlist.pickMap != null) {
				watchlist.pickMap.remove(security);
			}
		}
		super.delete(isAccountDeleting);

	}

	@Override
	public Strategy getOwner() {
		return watchlist;
	}

	public String getSimpleWatchlistLabel() {
		int index = watchlist.label.lastIndexOf(":");
		if (index == -1) {
			return watchlist.label;
		} else {
			return watchlist.label.substring(index + 1);
		}
	}

	public Date getStart() {
		return start;
	}

	public boolean isBuy() {
		return Signal.buy.name().equals(signal);

	}

	public boolean isHeld() {
		if (signal == null) {
			Log.warning("Why is my signal null?" + toString());
			return false;
		}

		return Signal.parse(signal).isHeld();
	}

	public boolean isOut() {
		return Signal.out.equals(signal);
	}

	public boolean isPending() {
		return Signal.parse(signal).isPending();
	}

	public boolean isSell() {
		return Signal.sell.name().equals(signal);

	}

	public boolean isWhitelisted() {
		return watchlist.whitelist.contains(security);
	}

	public Signal signal() {
		return Signal.parse(signal);
	}

	public void toBlacklisted() {
		Signal s = Signal.parse(signal);
		switch (s) {
		case out:
			return;
		case sell:
		case hold:
		case buy:
			becomeState(Signal.out, null);
			break;
		default:
			ExceptionUtil.throwUnsupported(s);
		}
	}

	@Override
	public String toString() {
		String accum = security + " pick for " + watchlist + "(" + start + ")";

		if (signal != null) {
			accum += "(" + signal + ")";
		}
		return accum;
	}

	public void toWhitelisted() {
		Signal s = Signal.parse(signal);
		switch (s) {
		case hold:
			return;
		case out:
		case sell:
		case buy:
			becomeState(Signal.hold, null);
			break;
		default:
			ExceptionUtil.throwUnsupported(s);
		}
	}

	/** Using signals as states */
	String becomeState(Signal newSignal, String label, String... reasons) {
		if (newSignal.equals(signal)) {
			return null;
		}
		String transition = signal + " " + newSignal.name();
		switch (transition) {

		case "buy hold":
			// Normal
			becomeHold();
			break;
		case "buy out":
			// Blacklist
			clearPending();
			break;
		case "buy sell":
			throw new UserException("Cannot sell " + toHandle() + " because you haven't bought it yet", false);

		case "hold sell":
			becomePending(newSignal, reasons);
			break;
		case "hold out":
			// blacklist can do this
			becomePending(Signal.sell, SignalReason.blacklist.name());
			becomeOut();
			break;
		case "hold buy":
			throw new UserException(toHandle() + " already holding", true);
		case "sell out":
			becomeOut();
			break;
		case "sell hold":
			clearPending();
			break;
		case "sell buy":
			throw new UserException(toHandle() + " already bought", false);
		case "out buy":
			becomePending(newSignal, reasons);
			break;
		case "out sell":
			throw new UserException(toHandle() + " already sold", true);
		case "out hold":
			becomePending(Signal.buy, SignalReason.whitelist.name());
			updateSignal(Signal.hold);
			break;

		default:
			throw new UserException(this + " should not go from " + signal + " to " + newSignal, false);
		}
		return null;
	}

	String processUserSell() {
		if (watchlist.whitelist.contains(security)) {
			watchlist.whitelist.remove(security);
			becomeOut();
			return label + " removed from whitelist";
		} else {
			watchlist.addToBlacklist(security);
			becomeOut();
			return label + " added to Blacklist";
		}

	}

	private void becomePending(Signal newSignal, String... reasons) {
		signalDate = new Date();
		if (newSignal.isEnter()) {
			signalStartDate = signalDate;
		} else {
			signalEndDate = signalDate;
		}
		signalCause = reasons;
		signal = newSignal.name();
	}

	private void updateSignal(Signal newSignal) {
		signal = newSignal.name();
		signalCause = null;
		signalDate = null;
	}

}
