package models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;

import javax.persistence.Entity;
import javax.persistence.ManyToMany;
import javax.persistence.OneToMany;
import javax.persistence.Transient;

import eye.play.control.PlayControl;
import eye.play.control.PlayUtil;
import eye.service.ServiceEnv;
import eye.service.stock.PositionService;
import eye.service.stock.Signal;
import eye.service.stock.SignalReason;
import eye.service.stock.TickerService;
import eye.service.stock.TickerService.Ticker;
import eye.transfer.EyeRecord;
import eye.transfer.EyeTable;
import eye.transfer.EyeType;
import eye.util.DateUtil;
import eye.util.ExceptionUtil;
import eye.util.ObjectUtil;
import eye.util.UserException;
import eye.util.logging.Log;

@Entity
public class Watchlist extends PickFilter {

	@ManyToMany
	public List<PickFilter> inFilters = new ArrayList();

	@ManyToMany
	public List<PickFilter> outFilters = new ArrayList();

	@OneToMany
	public PickFilter primary;

	public List<String> blacklist = new ArrayList();
	public List<String> whitelist = new ArrayList();

	@Transient
	public transient LinkedHashMap<String, StockPick> pickMap;
	public Date lastAlertUpdate;
	public boolean browsing;

	/**
	 * Some watchlists are linked to a brokerage account.
	 */
	public String brokerage;

	Watchlist() {
	}

	public static Date getScoreFromDate(StockPick pick) {
		Date bought = PlayUtil.watchlistScored;
		if (pick.isHeld()) {
			Date lastBought = pick.getStart();
			if (lastBought != null && lastBought.after(bought)) {
				bought = lastBought;
			}
		}
		return bought;
	}

	public void addToBlacklist(String sec) {
		if (blacklist.contains(sec) == false) {
			blacklist.add(sec);
		}
		// If it is on the whitelist, remove it
		whitelist.remove(sec);
	}

	public void addToWhitelist(String sec) {
		if (whitelist.contains(sec) == false) {
			whitelist.add(sec);
		}
		// If it is on the whitelist, remove it
		blacklist.remove(sec);
	}

	public void addToWhitelistBySymbol(String symbol) {
		Ticker ticker = TickerService.get().getTickerByName(symbol);
		addToWhitelist(ticker.getId());
	}

	//NOT using a setter as we only want to do this when the user changes the whitelist
	public void changeWhitelist(List<String> newWhitelist) {

		for (String item : newWhitelist) {
			if (whitelist.contains(item) == false) {
				StockPick pick = ensureStockPick(item);
				pick.toWhitelisted();
			}
		}
		whitelist = newWhitelist;
	}

	public StockPick createStockPick(String tickerId) {
		StockPick stockPick = new StockPick();
		stockPick.watchlist = this;
		stockPick.security = tickerId;
		stockPick.account = account;
		Ticker ticker = TickerService.get().getTickerById(tickerId);
		if (ticker == null) {
			throw new UserException(
					"Sorry, we cannot find the security with id " + tickerId + " in watchlist " + getLabel(), false);
		}
		stockPick.label = ticker.getLabel();
		stockPick.save();
		ensurePicks().put(tickerId, stockPick);
		return stockPick;

	}

	/**
	 * We do this regardless of whether we are saving or not...
	 */
	public EyeTable createTickerTable(EyeRecord buysAndSells) {
		EyeTable buysTable = buysAndSells.require("tickers");
		if (browsing) {
			buysTable.init();
			int actionIndex = buysTable.findColumn("action");
			for (Object[] row : buysTable.getRows()) {
				row[actionIndex] = Signal.hold.name();
			}
		}
		return buysTable;
	}

	@Override
	public void delete(boolean isAccountDeleting) {
		Collection<StockPick> picks = ensurePicks().values();
		ArrayList<StockPick> l = new ArrayList(picks);
		for (StockPick pick : l) {
			pick.delete(false);
		}
		//Watchlists are not referenced by anyone, so we don't have to check for references/slaves
		super.delete(true);
	}

	public HashMap<String, StockPick> ensurePicks() {
		if (pickMap == null) {
			pickMap = new LinkedHashMap();
			List<StockPick> list = PlayControl.stockPicks.createQuery().findByOwnerAndClause("watchlist", this,
					"target.end is NULL");
			for (StockPick pick : list) {
				StockPick dup = pickMap.get(pick.security);
				if (dup != null) {
					Log.warning(pick.security + " already exists in watchlist: " + this);
					merge(dup, pick);
					continue;
				}
				pickMap.put(pick.security, pick);
			}
		}

		return pickMap;
	}

	public StockPick ensureStockPick(String ticker) {
		StockPick stockPick = getStockPick(ticker);
		if (stockPick == null) {
			stockPick = createStockPick(ticker);
		}
		return stockPick;

	}

	public PickFilter findByName(String name, List<PickFilter> filters) {
		for (PickFilter p : filters) {
			if (p.getName().equals(name)) {
				return p;
			}
		}
		throw new IllegalStateException(name + " is not used by watchlist:" + filters);
	}

	public List<StockPick> getSignals() {
		List<StockPick> accum = new ArrayList();
		for (StockPick pick : ensurePicks().values()) {
			if (pick.isPending()) {
				accum.add(pick);
			}
		}
		return accum;
	}

	public StockPick getStockPick(String security) {
		StockPick pick = ensurePicks().get(security);
		return pick;

	}

	public boolean isImported() {
		return getLabel() != null && getLabel().startsWith("Imported");
	}

	public void merge(StockPick a, StockPick b) {
		assert a.end == null;
		assert b.end == null;
		if (a.start != null) {
			b.delete(false);
		} else {
			a.delete(false);
		}
	}

	/**
	 * Used for updates to sim brokerage.
	 */
	public EyeTable prepareAutoUpdateTable() {
		EyeTable t = new EyeTable();
		t.setName("watchlist-" + account.name + "-" + getId());
		t.addColumn("TickerId", EyeType.String);
		t.addColumn("Signal", EyeType.String);
		t.addColumn("Date", EyeType.Date);
		for (StockPick pick : ensurePicks().values()) {
			if (pick.isHeld() == false) {
				t.addRow(pick.security, pick.signal, pick.signalDate);
			}
		}
		return t;
	}

	public EyeTable prepareUpdateTable(boolean addStocks) {
		EyeTable t = new EyeTable();
		t.setName("tickers");
		t.addColumn("TickerId", EyeType.String);
		t.addColumn("active", EyeType.Integer);
		t.addColumn("score", EyeType.Float);
		t.addColumn("score_from", EyeType.Date);
		t.addColumn("PickId", EyeType.Long);
		if (addStocks) {
			for (StockPick pick : ensurePicks().values()) {
				int active = pick.isHeld() ? 2 : 1;

				Date bought = getScoreFromDate(pick);
				t.addRow(pick.security, active, 0, bought, pick.id);

			}
		}
		return t;
	}

	public void process() {
		List<StockPick> list = new ArrayList(ensurePicks().values());
		for (StockPick pick : list) {

			if (pick.isBuy()) {

			} else {
				//Not sure how we get into this case.
				whitelist.remove(pick.security);

			}
		}
	}

	public void processBuysAndSells(EyeRecord buysAndSells, StringBuilder userExceptions) {

		assert buysAndSells != null;
		// Might be more efficient to have henry to this.
		HashMap<String, StockPick> allPicks = new HashMap(ensurePicks());
		EyeTable t = buysAndSells.require("tickers");

		for (int i = 0; i < t.getRowCount(); i++) {
			try {
				String tickerId = (String) t.getValueAt(i, 0);
				Signal action = Signal.parse((String) t.getValueAt(i, 1));
				String[] reasons = (String[]) t.getValueAt(i, 2);
				if (action.isEnter()) {
					String reason = reasons != null ? reasons[0] : null;
					signalEnter(tickerId, reason, action);
					allPicks.remove(tickerId);
				} else if (action.isExit()) {
					signalExit(tickerId, reasons, action);
					allPicks.remove(tickerId);
				} else if (action == Signal.hold) {
					//In this case do not remove the pick, as we need to clear pending.
				} else {
					throw new IllegalStateException("What is:" + action);
				}
			} catch (UserException e) {
				userExceptions.append(e.getMessage());
				userExceptions.append("<br/>");
			}
		}

		for (String key : allPicks.keySet()) {
			StockPick pick = allPicks.get(key);

			if (pick.isPending()) {
				pick.clearPending();
			}

		}
		lastAlertUpdate = PlayControl.getEnv().date;

	}

	/** process signal as action. */
	public String processSignal(Signal signal, String tickerId, long pickId) {
		StockPick pick = getStockPick(tickerId);
		if (pick == null) {
			Ticker t = TickerService.get().getTickerById(tickerId);
			if (t == null || !t.getId().equals(tickerId)) {
				throw new UserException(tickerId + " no longer is a valid ticker", false);
			}
			throw new UserException("<HTML>" + t + " is no longer owned in " + this, true);
		}
		lastAlertUpdate = null;

		switch (signal) {
		case buy:
			pick.becomeState(Signal.hold, label);
			break;
		case sell:
			pick.becomeState(Signal.out, label);
			break;
		case neverBuy:
			pick.toBlacklisted();
			if (pick.isHeld()) {
				pick.becomeOut();
			}
			if (blacklist.contains(tickerId) == false) {
				blacklist.add(tickerId);
			}
			whitelist.remove(tickerId);
			break;
		case userSell:
			return pick.processUserSell();
		case whitelist:
			pick.clearPending();
			addToWhitelist(tickerId);
			break;
		default:
			throw ExceptionUtil.throwUnsupported(signal);
		}
		return null;
	}

	public StockPick requireStockPick(String security) {
		StockPick pick = getStockPick(security);
		if (pick == null) {
			Log.warning("missing " + security + " current are " + pickMap.keySet());
			throw new IllegalStateException("Why is " + security + " + not found?");
		}
		return pick;
	}

	public void resetPicks() {
		pickMap = null;

	}

	public void sellAllRemaining(HashMap<String, StockPick> picks) {
		List<StockPick> list = new ArrayList(picks.values());
		for (StockPick pick : list) {

			if (pick.isHeld()) {
				processSignal(Signal.userSell, pick.security, pick.id);
			} else {
				//Not sure how we get into this case.
				whitelist.remove(pick.security);

			}
		}

	}

	public void setInFilters(List<PickFilter> inFilters) {
		this.inFilters = inFilters;
	}

	public void setPicks(String type, List<String> tickers) {

		HashMap<String, StockPick> allPicks = (HashMap<String, StockPick>) ensurePicks().clone();

		for (String ticker : tickers) {

			StockPick pick = allPicks.remove(ticker);
			if (pick == null || pick.isOut()) {
				StockPick newPick = ensureStockPick(ticker);
				String[] reasons = { SignalReason.whitelist.name() };
				addToWhitelist(ticker);
				newPick.becomeState(Signal.hold, type, reasons);
			}
		}
		sellAllRemaining(allPicks);

	}

	public void setPrimary(PickFilter primary) {
		//if (!primary.equals(this.primary)) {
		this.primary = primary;
		if (tradingCost == 0) {
			tradingCost = primary.tradingCost;
		}
		if (maxHoldings == 0) {
			maxHoldings = primary.maxHoldings;
		}
		if (rebalance == null) {
			rebalance = primary.rebalance;
		}
		if (stopGain == null) {
			stopGain = primary.stopGain;
		}
		if (stopLoss == null) {
			stopLoss = primary.stopLoss;
		}
		if (tradingModel == null) {
			tradingModel = primary.tradingModel;
		}
		//}
	}

	public void testingSetPicks(String... tickerIds) {

	}

	public String toPrettyString() {
		StringBuilder b = new StringBuilder();
		b.append(toString());

		for (StockPick pick : ensurePicks().values()) {
			b.append("\n\t" + pick.security + "\t" + pick.signal + "\t:" + pick.start);
		}

		return b.toString();

	}

	public void updateFromPositions(EyeTable positions) {
		PositionService posService = new PositionService();
		posService.start(ServiceEnv.get());
		posService.loadPositions(positions);
		HashMap<String, StockPick> allPicks = new HashMap(ensurePicks());
		String label = "In auto-updating " + this + ":";
		StringBuilder accum = new StringBuilder();

		for (String key : allPicks.keySet()) {

			StockPick pick = allPicks.get(key);
			try {
				if (pick != null && pick.signalDate != null
						&& pick.signalDate.before(DateUtil.addDays(new Date(), 3))) {

					boolean isIn = posService.getPosition(pick.security) != null;
					if (isIn) {
						if (pick.isOut()) {
							accum.append(label + pick.toHandle() + " could not be sold. Why?\n");
						} else if (pick.isBuy()) {
							pick.becomeHold();
						}
					} else {
						if (pick.isSell()) {
							pick.becomeOut();
							//Give us three days to make sure that we have the correct starting price before warning of not bought..
						} else if (pick.isBuy()) {
							accum.append(label + pick.toHandle() + "signaled on " + ObjectUtil.format(pick.signalDate)
									+ " could not be bought. Why?\n");
						}

					}
				}

			} catch (Throwable e) {
				Log.severe("while processing " + pick.toHandle(), e);
			}
		}
		if (accum.length() > 0) {
			Log.warning(accum);
		}

	}

	private StockPick signalEnter(String security, String reason, Signal action) {
		StockPick pick = ensureStockPick(security);
		pick.becomeState(action, reason);
		return pick;
	}

	private void signalExit(String id, String[] reasons, Signal action) {
		StockPick pick = getStockPick(id);
		// Potentially not necessary, as it should be the case it exists...
		// Eventually, should put error checking here.
		if (pick != null) {
			pick.becomeState(action, null, reasons);
		}
	}

}
